@@ -10,9 +10,10 @@ import pendulum
 import python_graphql_client
 import pandas as pd
 import requests
+from time import sleep
 from datetime import datetime
 from typing import Optional, Dict, List, Union
-
+from tqdm.notebook import trange, tqdm
 from requests.auth import AuthBase
 from sumatra.auth import SDKKeyAuth, CognitoJwtAuth
 from sumatra.config import CONFIG
@@ -831,6 +832,8 @@ class Client:
         if "errors" in ret:
             raise Exception(ret["errors"][0]["message"])
 
+        return ret["data"]["deleteTimeline"]["id"]
+
     def infer_schema_from_timeline(self, timeline: str) -> str:
         """
         Attempt to infer the paths and data types of all fields in the timeline's
@@ -861,8 +864,57 @@ class Client:
 
         if "errors" in ret:
             raise Exception(ret["errors"][0]["message"])
+
         return ret["data"]["timeline"]["scowl"]
 
+    def create_timeline_from_s3(
+        self,
+        timeline: str,
+        s3_uri: str,
+        time_path: str,
+        data_path: str,
+        default_type: Optional[str] = None,
+        id_path: Optional[str] = None,
+        type_path: Optional[str] = None,
+    ):
+        query = """
+                    mutation SaveTimelineMutation($id: String!, $source: String!, $state: String!, $parameters: [KeyValueInput]!) {
+                        saveTimeline(id: $id, source: $source, state: $state, parameters: $parameters) {
+                            id
+                        }
+                    }
+                """
+
+        parameters = [
+            {"key": "s3_uri", "value": s3_uri},
+            {"key": "time_path", "value": time_path},
+            {"key": "data_path", "value": data_path},
+        ]
+
+        if default_type:
+            parameters.append({"key": "default_type", "value": default_type})
+
+        if id_path:
+            parameters.append({"key": "id_path", "value": id_path})
+
+        if type_path:
+            parameters.append({"key": "type_path", "value": type_path})
+
+        ret = self._gql_client.execute(
+            query=query,
+            variables={
+                "id": timeline,
+                "source": "s3",
+                "state": "processing",
+                "parameters": parameters,
+            },
+        )
+
+        if "errors" in ret:
+            raise Exception(ret["errors"][0]["message"])
+
+        return ret["data"]["saveTimeline"]["id"]
+
     def create_timeline_from_feed(
         self,
         timeline: str,
@@ -915,6 +967,58 @@ class Client:
             raise Exception(ret["errors"][0]["message"])
         self._wait_for_timeline_processing(timeline)
 
+    def create_timeline_from_log(
+        self,
+        timeline: str,
+        start_ts: Union[pd.Timestamp, str],
+        end_ts: Union[pd.Timestamp, str],
+        event_types: Optional[List[str]] = None,
+    ) -> None:
+        """
+        Create (or overwrite) timeline from the Event Log
+
+        Arguments:
+            timeline: Timeline name.
+            start_ts: Earliest event timestamp to fetch (local client timezone).
+            end_ts: Latest event timestamp to fetch (local client timezone).
+            event_types: Event types to include (default: all).
+        """
+
+        query = """
+            mutation SaveTimeline($id: String!, $parameters: [KeyValueInput]!) {
+              saveTimeline(id: $id, source: "athena", state: "processing", parameters: $parameters) {
+                id
+              }
+            }
+        """
+        if isinstance(start_ts, pd.Timestamp):
+            start_ts = str(start_ts)
+        if isinstance(end_ts, pd.Timestamp):
+            end_ts = str(end_ts)
+        start_ts = pendulum.parse(start_ts, tz=CONFIG.timezone).astimezone(
+            pendulum.timezone("UTC")
+        )
+        end_ts = pendulum.parse(end_ts, tz=CONFIG.timezone).astimezone(
+            pendulum.timezone("UTC")
+        )
+        start_str = start_ts.to_iso8601_string()
+        end_str = end_ts.to_iso8601_string()
+        parameters = [
+            {"key": "start", "value": start_str},
+            {"key": "end", "value": end_str},
+        ]
+
+        if event_types:
+            events = ",".join(event_types)
+            parameters.append({"key": "events", "value": events})
+
+        ret = self._gql_client.execute(
+            query=query, variables={"id": timeline, "parameters": parameters}
+        )
+        if "errors" in ret:
+            raise Exception(ret["errors"][0]["message"])
+        self._wait_for_timeline_processing(timeline)
+
     def create_timeline_from_dataframes(
         self,
         timeline: str,
@@ -1038,6 +1142,9 @@ class Client:
             jsonl += "\n"
         return jsonl
 
+    def get_materialization(self, id: str) -> Materialization:
+        return Materialization(self, id)
+
     def materialize(
         self, timeline: str, branch: Optional[str] = None
     ) -> Materialization:
@@ -1087,6 +1194,68 @@ class Client:
 
         return Materialization(self, ret["data"]["materialize"]["id"])
 
+    def distributed_materialize_many(
+        self,
+        timelines: List[str],
+        features: List[str] = None,
+        start_ts: Optional[Union[pd.Timestamp, str]] = None,
+        end_ts: Optional[Union[pd.Timestamp, str]] = None,
+        branch: Optional[str] = None,
+    ) -> Materialization:
+        """
+        Enrich collection of timelines using topology at branch. Timelines are merged based on timestamp.
+
+        This is the primary function of the SDK.
+
+        Arguments:
+            timelines: Timeline names.
+            branch: Specify a source branch other than the client default.
+            start_ts: Earliest event timestamp to materialize (local client timezone).
+            end_ts: Latest event timestamp to materialize (local client timezone).
+            features: List of features to materialize, e.g. `['login.email', 'purchase.*']`
+
+        Returns:
+            Handle to Materialization job
+        """
+
+        variables = {
+            "timelines": timelines,
+            "branch": branch or self._branch,
+            "features": features,
+        }
+
+        if start_ts:
+            if isinstance(start_ts, pd.Timestamp):
+                start_ts = str(start_ts)
+            start_ts = pendulum.parse(start_ts, tz=CONFIG.timezone).astimezone(
+                pendulum.timezone("UTC")
+            )
+            variables["start"] = start_ts.to_iso8601_string()
+
+        if end_ts:
+            if isinstance(end_ts, pd.Timestamp):
+                end_ts = str(end_ts)
+            end_ts = pendulum.parse(end_ts, tz=CONFIG.timezone).astimezone(
+                pendulum.timezone("UTC")
+            )
+            variables["end"] = end_ts.to_iso8601_string()
+
+        query = """
+            mutation DistributedMaterialize($timelines: [String], $branch: String!, $features: [String], $start: DateTime, $end: DateTime) {
+                distributedMaterialize(timelines: $timelines, branch: $branch, features: $features, start: $start, end: $end) { id }
+            }        
+        """
+
+        ret = self._gql_client.execute(
+            query=query,
+            variables=variables,
+        )
+
+        if "errors" in ret:
+            raise Exception(ret["errors"][0]["message"])
+
+        return Materialization(self, ret["data"]["distributedMaterialize"]["id"])
+
     def get_models(self) -> pd.DataFrame:
         """
         Return all PMML models and their metadata.
@@ -1213,7 +1382,7 @@ class Materialization:
     def _get_materialization(self):
         query = """
             query Materialization($id: String!) {
-                materialization(id: $id) { id, timelines, branch, hash, state, path }
+                materialization(id: $id) { id, state, path, jobsSubmitted, jobsCompleted }
             }
         """
 
@@ -1244,6 +1413,56 @@ class Materialization:
         self._mtr = self._get_materialization()
         return self._mtr["state"]
 
+    def wait(self) -> str:
+        """
+        Wait until materialization completes. In a Notebook, a progress bar is displayed.
+
+        Returns:
+            Materialization status
+        """
+        self._mtr = self._get_materialization()
+        if self._mtr["state"] != "processing":
+            return self._mtr["state"]
+
+        while self._mtr["jobsSubmitted"] == 0 and self._mtr["state"] == "processing":
+            sleep(0.5)
+            self._mtr = self._get_materialization()
+
+        submitted = self._mtr["jobsSubmitted"]
+        completed = self._mtr["jobsCompleted"]
+
+        with tqdm(total=submitted) as pbar:
+            pbar.update(completed)
+            while self._mtr["state"] == "processing":
+                sleep(0.5)
+                self._mtr = self._get_materialization()
+                new_completed = self._mtr["jobsCompleted"]
+                new_submitted = self._mtr["jobsSubmitted"]
+                if new_submitted != submitted:
+                    pbar.total = new_submitted
+                    submitted = new_submitted
+
+                pbar.update(new_completed - completed)
+                completed = new_completed
+            if self._mtr["state"] == "materialized":
+                pbar.update(submitted - completed)
+
+        return self._mtr["state"]
+
+    @property
+    def progress(self) -> str:
+        """
+        Current progress of subjobs: X of Y jobs completed.
+        """
+        self._mtr = self._get_materialization()
+
+        if self._mtr["state"] != "processing":
+            return self._mtr["state"]
+        else:
+            completed = self._mtr["jobsCompleted"]
+            submitted = self._mtr["jobsSubmitted"]
+            return f"{completed} / {submitted} jobs completed."
+
     @property
     def timelines(self) -> List[str]:
         """
@@ -1304,24 +1523,26 @@ class Materialization:
         Returns:
             Enriched events
         """
-        if event_type not in self.event_types:
-            raise Exception(f"Event type '{event_type}' not found.")
+        # if event_type not in self.event_types:
+        #    raise Exception(f"Event type '{event_type}' not found.")
 
         self._wait_for_processing()
         session = self._client._get_session()
 
+        path = f"{self.path}/events/event_type={event_type}/"
+        cols = ["_id", "_type", "_time"]
+
         if not features:
             df = wr.s3.read_parquet(
                 boto3_session=session,
-                path=f"{self.path}/{event_type}.parquet",
+                path=path,
                 use_threads=8,
             )
         else:
-            cols = ["_id", "_type", "_time"]
             cols.extend(features)
             df = wr.s3.read_parquet(
                 boto3_session=session,
-                path=f"{self.path}/{event_type}.parquet",
+                path=path,
                 columns=cols,
                 use_threads=8,
             )
@@ -1341,14 +1562,13 @@ class Materialization:
         Returns:
             Event-level errors
         """
-
-        if event_type not in self.event_types:
-            raise Exception(f"Event type '{event_type}' not found.")
+        # if event_type not in self.event_types:
+        #    raise Exception(f"Event type '{event_type}' not found.")
 
         self._wait_for_processing()
         session = self._client._get_session()
 
-        path = f"{self.path}/{event_type}.errors.parquet"
+        path = f"{self.path}/errors/event_type={event_type}/"
         cols = ["_id", "_type", "_time"]
         cols.extend(features)
 
